import time
import threading
import warnings
from typing import Optional, NoReturn, Union

import requests

from .abstarct.stress_test import AbstractStressTest


class StressTest(AbstractStressTest):
    """
    Класс для стресс-тестирования интернет-ресурсов.
    (с помощью отправки большого кол-ва запросов и проверки,
    через сколько ресурс перестанет отвечать).
    Имеется возможность ручного и автоматического создания потоков для отправки запросов.
    Есть встроенная автоматическая и ручная статистика.
    """

    def __init__(self, host: str = '127.0.0.1'):
        """
        :param host: str. Хост куда будут отправляться запросы
        :param _func_increase_connections: Функция отвечающая за кол-во потоков,
        которое необходимо создать. По умолчанию + 5%
        :param _kill: bool. При значении True потоки перестанут отправлять запросы
        :param _thread: list. Список с потоками (threading.Thread)
        :param _stats: dict. Словарь, в котором:
        key - Время добавления в словарь
        value - За сколько ресурс обработал запрос (время)
        :param _errors: list. Список с ошибками.
        При использовании auto_get_stats, будет отчищаться после каждого успешного запроса.
        [0] - Время ошибки
        [1] - Список с потоками (self._thread)
        [2] - Список с запросами (self._stats)
        :param _timeout: int. Задержка. Используется:
        1) При создании потоков;
        2) При отправке запросов.
        :param max_thread_count: int. Максимальное кол-во создаваемых потоков (не менее 20).
        Реальное кол-во потоков может быть незначительно больше указанного.
        """
        self._host = host
        self._func_increase_connections = lambda count_con: int(count_con / 100 * 5)
        self._kill = False
        self._thread = []
        self._stats = {}
        self._errors = []
        self._timeout = 3
        self._max_thread_count = 99999
        self._time_start = 0
        self._previous_count_errors = 0
        self._max_execution_time = 60 * 60 * 24 * 31  # 1 месяц

    def auto_create_connection(self, timeout: Optional[Union[int, float]] = None) -> None:
        """
        Автоматическое создание соединений к ресурсу.
        :return: None
        """
        self._time_start = time.time()
        th = threading.Thread(target=self._auto_create_connection, args=(timeout or self.timeout,))
        th.start()

    def auto_get_stats(self, timeout: Union[int, float] = 5) -> Union[list, int]:
        """
        Метод для автоматического получения статистики по работе класса.
        В случае убийства всех запросов или возникновения ошибки, метод сообщит об этом
        с помощью warnings.warn.
        Если произошло событие (был отправлен запрос), вернёт stats с последними событиями
        (см. в __init__ расшифровку stats).
        Время задержки перед каждыми сбором статистики определяется передаваемым timeout.
        :param max_thread_count: int (максимальное кол-во поток, приоритетнее чем self._max_thread_count)
        :param timeout: Union[int, float]
        :return: list
        """
        previous_count_connection = len(self.stats)
        while True:
            if self.kill:
                warnings.warn('Все соединения убиты!')
                break
            if time.time() > self._time_start + self._max_execution_time:
                print(f'Время выполнения теста закончилось.\nВыполнялся: {self._max_execution_time} секунд.')
                self.kill_all_connections()
            error_quantity = self._get_last_number_errors()
            if error_quantity:
                warnings.warn(f"Прозошла(и) {error_quantity} ошибок!")
                yield error_quantity
            if previous_count_connection == len(self.stats):
                continue
            yield list(self._stats.items())[previous_count_connection:]
            previous_count_connection = len(self.stats)
            time.sleep(timeout)

    def create_connection(self, connection_type: str = 'GET') -> None:
        """
        Используется для создания соединения.
        Можно создавать несколько одновременных соединений, так как используются потоки.
        При вызове автоматически добавляет поток в список threads.
        :param connection_type:
        :return:
        """
        th = threading.Thread(target=self._start_connection_stress_test, args=(connection_type,))
        self._thread.append(th)
        th.start()

    def set_increase_connections(self, func) -> None:
        """
        Устанавливает функцию для увелечения кол-ва потоков.
        :param func: function (Любая функция принимающая 1 параметр
        (текущее кол-во потоков и возвращающая число))
        :return: int (Число потоков, которое необходимо создать, 0 - допускается)
        """
        self._func_increase_connections = func

    def kill_all_connections(self) -> None:
        """
        Закрывает все соединения.
        :return: None
        """
        self._kill = True

    def _start_connection_stress_test(self, connection_type: str) -> None:
        """
        Отправка запросов, с задержкой указанной в self._timeout.
        Если соединения убиты или словарь self._errors содержит более 250 ошибок,
        отправка запросов прекращается.
        Добавляет запрос в self._errors, если возникла ConnectionError
        или код ответа от сервера != 200.
        :param connection_type: str (тип запроса 'GET', 'POST' и т.д.)
        :return: None
        """
        while not self._kill:
            func = requests.request
            kwargs = {'method': connection_type, 'url': self._host}
            time_now = time.time()
            try:
                response, runtime = self._get_runtime(func, **kwargs)
            except requests.ConnectionError:
                self._errors.append([time.time(), len(self._thread), len(self._stats)])
                continue
            self._stats[time_now - runtime] = runtime
            self._check_response_status_code(response)
            time.sleep(self._timeout)

    def _check_response_status_code(self, response) -> None:
        """
        Если код ответа != 200, записывает в self._errors
        (см. в __init__, что записывается в self._errors).
        :param response:
        :return:
        """
        if response.status_code != 200:
            self._errors.append([time.time(), len(self._thread), len(self._stats)])

    def _auto_create_connection(self, timeout: int) -> None:
        """
        Автоматически создаёт 20 потоков,
        после чего создаёт их по правилу функции self._func_increase_connections.
        Если потоки убиты (self._kill = True), прекращает созание потоков
        :return: None
        """
        for _ in range(20):
            self.create_connection()
        while not self._kill:
            if len(self.thread) <= self.max_thread_count:
                for _ in range(self._func_increase_connections(len(self._thread))):
                    self.create_connection()
            time.sleep(timeout)

    @staticmethod
    def _get_runtime(func, *args, **kwargs) -> tuple:
        """
        Возвращает результат запроса и время, за которое этот запрос был выполнен.
        :param func: Любая функция (она и будет запускаться).
        :param args: Аргументы передаваемые в функцию.
        :param kwargs: Именованные аргументы передаваемые в функцию.
        :return: tuple (первый параметр - результат выполнения функции,
        второй параметр - время выполнения в секундах)
        """
        time_start = time.time()
        response = func(*args, **kwargs)
        return response, time.time() - time_start

    def _get_last_number_errors(self) -> int:
        """
        Получить последние ошибки.
        :return: None
        """
        result = len(self._errors) - self._previous_count_errors
        self._previous_count_errors = len(self._errors)
        return result

    @property
    def thread(self) -> list:
        return self._thread

    @property
    def stats(self) -> dict:
        return self._stats

    @property
    def timeout(self) -> int:
        return self._timeout

    @timeout.setter
    def timeout(self, value: Union[int, float]) -> Optional[NoReturn]:
        """
        Устанавливает переменную _timeout.
        Значение должны быть int|float и > 0.
        :param value: Union[int, float]
        :return: None
        """
        if not isinstance(value, (int, float)) or value != abs(value):
            raise ValueError('Задержка должна быть числом (int, float) > 0')
        self._timeout = value

    @property
    def max_execution_time(self) -> Union[int, float]:
        return self._max_execution_time

    @max_execution_time.setter
    def max_execution_time(self, value: Union[int, float]) -> None:
        """
        Устанавливает максимальное время тестирования.
        :param value: Union[int, float](value >= 0)
        :return: None
        """
        if not isinstance(value, (int, float)) or value <= 0:
            raise ValueError('Значение max_execution_time может быть только типа int, float > 0')
        self._max_execution_time = value

    @property
    def max_thread_count(self):
        return self._max_thread_count

    @max_thread_count.setter
    def max_thread_count(self, value: int):
        """
        Устанавливает максимальное кол-во потоков.
        :param value: int (value > 20)
        :return: None
        """
        if not isinstance(value, int) or value != abs(value) or value < 20:
            raise ValueError('Максимальное кол-во потоков должно быть числом (int) > 20')
        self._max_thread_count = value

    @property
    def kill(self):
        return self._kill

    @kill.setter
    def kill(self, value: bool) -> Optional[NoReturn]:
        """
        Устанавливает параметр _kill.
        Если = True, запросы перестанут посылаться.
        :param value: bool
        :return: None
        """
        if not isinstance(value, bool):
            raise ValueError('Значение kill может быть только типа bool')
        self._kill = value

    @property
    def errors(self):
        return self._errors



from abc import ABC, abstractmethod
from typing import Union, Optional


class AbstractMonitor(ABC):
    @abstractmethod
    def start(self):
        pass

    @abstractmethod
    def set_params(self, kill: Optional[bool], timeout: Optional[Union[float, int]], max_thread_count: Optional[int]):
        pass

    @abstractmethod
    def build_graph(self):
        pass

    @abstractmethod
    def stop(self):
        pass



import time
from statistics import mean
from typing import Optional, Union, Tuple

from matplotlib import pyplot as plt

from .abstarct.monitor import AbstractMonitor
from .abstarct.stress_test import AbstractStressTest


class BaseMonitor(AbstractMonitor):
    def __init__(self, stress: AbstractStressTest):
        self._stress = stress

    def start(self, timeout: Union[int, float] = 0.2) -> None:
        """
        Запускает тест.
        :param timeout: Union[int, float] (время задержки, используется при создании потоков и отправки запросов)
        :return: None
        """
        self._stress.auto_create_connection()
        for _ in self._stress.auto_get_stats(timeout):
            print(f'Всего запросов: {len(self._stress.stats)}')

    def set_params(self, kill: Optional[bool] = None, timeout: Optional[Union[float, int]] = None,
                   max_thread_count: Optional[int] = None, max_execution_time: Optional[int] = None) -> None:
        """
        Устанавливает параметры стресс-теста.
        :param kill: bool (True - убить потоки, False - продолжить тестирование)
        :param timeout: Union[int, float] (время задержки, используется при создании потоков и отправки запросов)
        :param max_thread_count: int (максимальное кол-во потоков,
        реальное кол-во может незначительно отличаться от заданного)
        :param max_execution_time: int (максимальное время выполнения теста)
        :return: None
        """
        self._stress.kill = kill or self._stress.kill
        self._stress.timeout = timeout or self._stress.timeout
        self._stress.max_thread_count = max_thread_count or self._stress.max_thread_count
        self._stress.max_execution_time = max_execution_time or self._stress.max_execution_time

    def stop(self) -> None:
        """
        Останавливает работу теста и запускает построение статистики.
        :return: None
        """
        self._stress.kill_all_connections()
        self.build_graph()


class ChartMonitor(BaseMonitor):
    def build_graph(self) -> None:
        """
        Построить статистику в виде графика. Используется matplotlib.
        Первый график - зависимость времени выполнения запроса от кол-ва запросов.
        Второй график - зависимость кол-ва ошибок от кол-ва запросов.
        :return: None
        """
        stats = sorted(self._stress.stats.items(), key=lambda i: i[0])
        x_time_request = list(map(lambda i: round(time.time() - i[0], 4), stats))
        y_time_request = list(map(lambda i: round(i[1], 4), stats))
        self._draw_plot(x_time_request, y_time_request, [2, 1, 1], 'Время отправки запроса')
        plt.ylabel('Время выполнения запроса')

        self._draw_plot(*self._get_x_y_error_coord(), [2, 1, 2], color='r')
        plt.xlabel('Время ошибки')
        plt.ylabel('Кол-во ошибок')
        plt.show()

    def _get_x_y_error_coord(self) -> Tuple[list, list]:
        """
        Получение кортежа с координатами ошибок x, y.
        tuple[0] - list (координата x)
        tuple[1] - list (координата y)
        В качестве координат x, идёт время, когда проищошла ошибка.
        В качестве координат y, идёт кол-во ошибок, которое проищошло за время указанное в x (по умолчанию за 1 секунду)
        :return: Tuple[list, list]
        """
        res = [[]]
        x_error = y_error = []
        if len(self._stress.errors) > 0:
            back = int(self._stress.errors[0][0])
            # Группируем ошибки. Ошибки за одну секунду будут в своём списке
            for i in self._stress.errors:
                i[0] = int(i[0])
                # Если время ошибки совпадает с предыдущим, добавляем в один список
                if i[0] == back:
                    res[-1].append(i)
                # Иначе создаём список с новым временем ошибки
                else:
                    res.append([i])
                    back = i[0]
            # Получаем время когда проихошла ошибка
            x_error = list(map(lambda el: round(time.time() - el, 4), [i[0][0] for i in res]))
            # Получаем кол-во ошибок
            y_error = [len(i) for i in res]
        return x_error, y_error

    def _draw_plot(self, x: list, y: list, coord: Optional[list] = None, title: Optional[str] = None,
                   color: str = 'b') -> None:
        """
        Строит график по указанным координатам x, y.
        :param x: list (координаты x (работает при условии len(x) == len(y)))
        :param y: list (координаты y (работает при условии len(x) == len(y)))
        :param coord: Optional[list] (координаты разбивания графика (matplotlib.subplot))
        :param title: Optional[str] (название графика (matplotlib.title))
        :param color: str (цвет графика)
        :return: None
        """
        if coord:
            plt.subplot(*coord)
        plt.plot(x, y, c=color)
        if title:
            plt.title(title)
        plt.ticklabel_format(useOffset=False)


class CMDMonitor(BaseMonitor):
    def start(self, timeout: Union[int, float] = 0.2) -> None:
        """
        Запускает тест, отобрадение статистики происходит в терминале.
        :param timeout: Union[int, float] (время задержки, используется при создании потоков и отправки запросов)
        :return: None
        """
        self._stress.auto_create_connection()
        for i in self._stress.auto_get_stats(0.2):
            if isinstance(i, int):
                continue
            print(f'Потоков: {len(self._stress.thread)}')
            print(f'Последние запросы: {len(i)}')
            print(f'Запросов всего: {len(self._stress.stats)}')
            print(f'Ошибок: {len(self._stress.errors)}', end='\n\n\n')

    def build_graph(self) -> None:
        """
        Строит график в терминале.
        Параметры вывода:
        1) Всего потоков
        2) Всего запросов
        3) Среднее время на запрос
        4) Всего ошибок
        :return: None
        """
        stats = self._stress.stats.items()
        average_time_per_request = mean([i[1] for i in stats])
        print(f'\n\n\nВсего потоков: {len(self._stress.thread)}')
        print(f'Всего запросов: {len(stats)}')
        print(f'Среднее время на запрос: {average_time_per_request}')
        print(f'Всего ошибок: {len(self._stress.errors)}', end='\n\n\n')



from abc import ABC, abstractmethod
from typing import Union, Optional


class AbstractStressTest(ABC):
    @abstractmethod
    def auto_create_connection(self, timeout: Optional[Union[int, float]] = None) -> None:
        pass

    @abstractmethod
    def auto_get_stats(self, timeout: Union[int, float] = 5) -> Union[list, int]:
        pass

    @abstractmethod
    def create_connection(self, connection_type: str = 'GET') -> None:
        pass

    @abstractmethod
    def kill_all_connections(self) -> None:
        pass

    @abstractmethod
    def set_increase_connections(self, func) -> None:
        pass

    @property
    def stats(self) -> dict:
        pass

    @property
    def errors(self) -> dict:
        pass

    @property
    def thread(self) -> dict:
        pass



from setuptools import setup, find_packages

with open('README.md', 'r', encoding='utf-8') as f:
    LONG_DESCRIPTION = f.read()

setup(
    name='pystresstest',
    version='1.1.2',
    description='Module for resource stress testing and automatic statistics generation',
    long_description=LONG_DESCRIPTION,
    license="MIT",
    author='Chekashov Matvey/Ryize',
    author_email='chekashovmatvey@gmail.com',
    url="https://github.com/Ryize/pyStressTest",
    packages=find_packages(),
    install_requires=['requests', 'matplotlib']
)



"""
Здесь вы можете написать свой код для стресс проверки ресурса.
Пример программы:

if __name__ == '__main__':
    stress = StressTest('http://127.0.0.1:8000')
    stress.timeout = 0.5
    stress.max_execution_time = 60
    stress.max_thread_count = 100
    monitor = ChartMonitor(stress)
    monitor.start()
    monitor.build_graph()
"""

if __name__ == '__main__':
    from stress_test import StressTest
    from monitor import ChartMonitor

    stress = StressTest('http://127.0.0.1:8000')
    stress.timeout = 0.5
    stress.max_execution_time = 10
    stress.max_thread_count = 100
    monitor = ChartMonitor(stress)
    monitor.start()
    monitor.build_graph()
